version = '1.5.0'
array_module = 'numpy'
array_module_version = '1.11.3'
python_version = '2.7.1'
formats = {}
formats['hdfeos5'] = 0
formats['netcdf4'] = 1
formats['hdf4'] = 1
formats['hdf5'] = 1
formats['shapefile'] = 1
formats['grib2'] = 1
formats['hdfeos'] = 1
formats['opendap'] = 1
